import argparse
import logging
import bibtexparser
from os import path, listdir
import pathlib
import os
import pdf2doi
from argparse import RawTextHelpFormatter
import pkgutil
import pdfrenamer.config as config
from pdfrenamer.filename_creators import build_filename, find_tags_in_format, AllowedTags 

#Change the formatter for the logger of the pdf2doi library, in order to add a prefix in front of all
#output created by pdf2doi
logger = logging.getLogger("pdf2doi")
logger.setLevel(level=logging.INFO)
logger.handlers =[]
if not logger.handlers:
    formatter = logging.Formatter("\t[pdf2doi]: %(message)s")
    ch = logging.StreamHandler()
    ch.setFormatter(formatter)
    logger.addHandler(ch)
logger.propagate = False
#####

def rename(target, verbose=False, format=config.format, 
           max_length_authors=config.max_length_authors, max_length_filename=config.max_length_filename,
           check_subfolders = False,
           tags=None):
    '''
    This is the main routine of the script
    Parameters
    ----------
    target : string
        Relative or absolute path of the target .pdf file or directory
    verbose : boolean, optional
        Increases the output verbosity. The default is False.
    format : string, optional
        Specifies the format of the filename. If not specified, the default value config.format is used instead.
    max_length_authors : integer, optional
        Sets the maximum length of any string related to authors. If not specified, the default value config.max_length_authors is used instead.
    max_length_filename : integer, optional
        Sets the maximum length of any generated filename. Any filename longer than this will be truncated. If not specified, 
        the default value config.max_length_filename is used instead.
    check_subfolders : boolean, optional
        If set true, and if target is a directory, all sub-directories will be scanned for pdf files to be renamed. Default value is False.

    Returns
    -------
    results, dictionary or list of dictionaries (or None if an error occured)
        The output is a single dictionary if target is a file, or a list of dictionaries if target is a directory, 
        each element of the list describing one file. Each dictionary has the following keys
        result['path_original'] = path of the pdf file (with the original filename)
        result['path_new'] = path of the pdf file (with the new filename)
        result['identifier'] = DOI or other identifier (or None if nothing is found)
        result['identifier_type'] = string specifying the type of identifier (e.g. 'doi' or 'arxiv')
        result['validation_info'] = Additional info on the paper possibly returned by the pdf2doi library
        result['method'] = method used to find the identifier

    '''
    
    #config.numb_results_google_search = numb_results_google_search

    # Setup logging
    if verbose: loglevel = logging.INFO
    else: loglevel = logging.CRITICAL

    logger = logging.getLogger("pdf-renamer")
    logger.setLevel(level=loglevel)
    
    #Make some sanity check on the format
    if not format:
        logger.error(f"The specified format is not a valid string.")
        return None
    tags = find_tags_in_format(format)
    if not tags:
        logger.error(f"The specified format does not contain any tag.")
        return None
    for tag in tags:
        if not tag in AllowedTags:
            logger.error(f"The specified format contains \"{tag}\", which is not a valid tag.")
            logger.error(f"The valid tags are: " + ",".join(AllowedTags))
            return None
    
    #Check if target is a directory
        # If yes, we look for all the .pdf files inside it, and for each of them
        # we call again this function by passing the file path as target.
        #
        # Moreover, if check_subfolders==True, for each subfolder in the directory we call again this function
        # by passsing the subfolder as target

    if  path.isdir(target):
        logger.info(f"Looking for pdf files and subfolders in the folder {target}...")
        if not(target.endswith(config.separator)): #Make sure the path ends with "\" or "/" (according to the OS)
                target = target + config.separator
        #Check if a file "journal_abbreviations.txt" exists in the folder. If yes, we use it as additional source of abbreviations
        if path.exists(target + "journal_abbreviations.txt"):
            logger.info(f"Found a file journal_abbreviations.txt with possible additional Journal abbreviations.")
            config.additional_abbreviations_file = target + "journal_abbreviations.txt"

        #We build a list of all the pdf files in this folder, and of all subfolders
        pdf_files = [f for f in listdir(target) if f.endswith('.pdf')]
        subfolders = [ f.path for f in os.scandir(target) if f.is_dir() ]

        numb_files = len(pdf_files)
        if numb_files == 0:
            logger.error("No pdf file found in this folder.")
        else:
            logger.info(f"Found {numb_files} pdf file(s).")

            
            files_processed = [] #For each pdf file in the target folder we will store a dictionary inside this list
            for f in pdf_files:
                logger.info(f"................") 
                file = target + f
                #We call again this same function but this time targeting the single file
                result = rename(file, verbose=verbose, format=format, 
                                max_length_filename=max_length_filename, max_length_authors= max_length_authors, 
                                tags=tags)
                files_processed.append(result)
            logger.info("................") 

        #If there are subfolders, and if check_subfolders==True, we call gain this function for each subfolder
        numb_subfolders = len(subfolders)
        if numb_subfolders:
            logger.info(f"Found {numb_subfolders} subfolder(s)")
            if check_subfolders==True :
                logger.info("Exploring subfolders...") 
                for subfolder in subfolders:
                    result = rename(subfolder, verbose=verbose, format=format, 
                                    max_length_filename=max_length_filename, max_length_authors= max_length_authors, 
                                    check_subfolders=True, tags=tags)
                    files_processed.extend(result)
            else:
                logger.info("The subfolder(s) will not be scanned because the parameter check_subfolders is set to False."+
                            " When using this script from command line, use the option -sf to explore also subfolders.") 


        return files_processed
    
    #If target is not a directory, we check that it is an existing file and that it ends with .pdf
    else:
        filename = target
        logger.info(f"File: {filename}")  
        if not path.exists(filename):
            logger.error(f"'{filename}' is not a valid file or directory.")
            return None    
        if not filename.endswith('.pdf'):
            logger.error("The file must have .pdf extension.")
            return None
        
        #We use the pdf2doi library to retrieve the identifier and info of this file
        logger.info(f"Calling the pdf2doi library to retrieve identifier and info of this file.")
        result = pdf2doi.pdf2doi(   filename ,verbose=verbose,
                                    websearch=True, webvalidation=True,
                                    numb_results_google_search=config.numb_results_google_search)
        result['path_original'] = filename
        if result and result['identifier']:
            logger.info(f"Found an identifier for this file: {result['identifier']} ({result['identifier_type']}).")
            data = result['validation_info']
            data = bibtexparser.loads(data)
            metadata = data.entries[0]
            #metadata_string = "\t\t\t\t"+"\n\t\t\t\t".join([f"{key} = \"{metadata[key]}\"" for key in metadata.keys()] ) 
            #logger.info("Found the following info:")
            #logger.info(metadata_string)
            NewName = build_filename(metadata, format, tags, max_length_filename=max_length_filename, max_length_authors= max_length_authors)
            ext = os.path.splitext(filename)[-1].lower()
            directory = pathlib.Path(filename).parent
            NewPath = str(directory) + config.separator + NewName
            NewPathWithExt = str(directory) + config.separator + NewName + ext
            logger.info(f"The new filename is {NewPathWithExt}")
            if (filename==NewPathWithExt):
                logger.info("The new filename is identical to the old one. Nothing will be changed")
                result['path_new'] = NewPathWithExt
            else:
                try:
                    NewPathWithExt_renamed = rename_file(filename,NewPath,ext) 
                    logger.info(f"File renamed correctly.")
                    if not (NewPathWithExt == NewPathWithExt_renamed):
                        logger.info(f"(Note: Another file with the same name was already present in the same folder, so a numerical index was added in the end).")
                    result['path_new'] = NewPathWithExt_renamed
                except Exception as e: 
                    logger.error('Some error occured while trying to rename this file: \n '+ str(e))
                    result['path_new'] = None
        else:
            logger.info("The pdf2doi library was not able to find an identifier for this pdf file.")
            result['path_new'] = None
        
        return result 

def rename_file(old_path,new_path,ext):
    #It renames the file in old_path with the new name contained in new_path. 
    #If another file with the same name specified by new_path already exists in the same folder, it adds an 
    #incremental number (e.g. "filename.pdf" becomes "filename (2).pdf"

    if not os.path.exists(old_path):
        raise ValueError(f"The file {old_path} does not exist")
    i=1

    while True:
        New_path = new_path + (f" ({i})" if i>1 else "") + ext
        if os.path.exists(New_path):
            i = i+1
            continue
        else:
            os.rename(old_path,New_path)
            return New_path

def main():
    parser = argparse.ArgumentParser( 
                                    description = "Automatically renames pdf files of scientific publications by retrieving their identifiers (e.g. DOI or arxiv ID) and looking up their bibtex infos.",
                                    epilog = "",
                                    formatter_class=RawTextHelpFormatter)
    parser.add_argument(
                        "path",
                        help = "Relative path of the pdf file or of a folder.",
                        metavar = "path")
    parser.add_argument(
                        "-nv",
                        "--no_verbose",
                        help="Decrease verbosity of output.",
                        action="store_true")
    parser.add_argument('-f', 
                        help="Format of the new filename. Default = \"{YYYY} - {Jabbr} - {A3etal} - {T}\".\n"+
                        "Valid tags:\n"+
                        "\n".join([key+val for key,val in AllowedTags.items()]),
                        action="store", dest="format", type=str, default = "{YYYY} - {Jabbr} - {A3etal} - {T}")
    parser.add_argument(
                        "-sf",
                        "--sub_folders",
                        help="Rename also pdf files contained in subfolders of target folder.",
                        action="store_true")
    parser.add_argument('-max_length_authors', 
                        help=f"Sets the maximum length of any string related to authors (default={str(config.max_length_authors)}).",
                        action="store", dest="max_length_authors", type=int)
    parser.add_argument('-max_length_filename', 
                        help=f"Sets the maximum length of any generated filename. Any filename longer than this will be truncated (default={str(config.max_length_filename)}).",
                        action="store", dest="max_length_filename", type=int)
    parser.add_argument('-google_results', 
                        help=f"Set how many results should be considered when doing a google search for the paper identifier via the pdf2doi library (default={str(config.numb_results_google_search)}).",
                        action="store", dest="google_results", type=int)

    
    args = parser.parse_args()
    results = rename(target=args.path,
                  verbose=not(args.no_verbose),
                  format=args.format,
                  max_length_authors = args.max_length_authors,
                  max_length_filename = args.max_length_filename,
                  check_subfolders = args.sub_folders
                  )

    print("Summaries of changes done:")    
    if not results:
        print("No file has been renamed.")
        return

    if not isinstance(results,list):
        results = [results]
    
    counter = 0
    for result in results:
        if result['identifier'] and not(result['path_original']==result['path_new']):
            print(f"\'{result['path_original']}\' --> \'{result['path_new']}\'")
            counter = counter + 1
    if counter==0:
        print("No file has been renamed.")
    else:
        string = f"{counter} file" + ("s have " if counter>1 else " has ") + "been renamed."
        print(string)

    return

if __name__ == '__main__':
    main()


''' 
This module contains several functions and variables that are used to generate
a valid filename, based on the chosen format and the available infos
'''

import re
import pkgutil
import pdfrenamer.config as config
import logging
logger = logging.getLogger("pdf_renamer")

AllowedTags = {"{YYYY}":" \t\t=\t Year of publication",
                "{MM}":" \t\t=\t Month of publication (in digits)",
                "{DD}":" \t\t=\t Day of publication (in digits)",
                "{J}":" \t\t=\t Full name of Journal",
                "{Jabbr}":" \t=\t Abbreviated name of Journal (if available)",
                "{Aall}":" \t\t=\t Last name of all authors (separated by comma)",
                "{Aetal}":" \t=\t Last name of the first author, add \'et al.\' if more authors are present",
                "{A3etal}":" \t=\t Last name of the first three authors (separated by comma), add \'et al.\' if more authors are present",
                "{aAall}":" \t=\t First initial and last name of all authors (separated by comma)",
                "{aAetal}":" \t=\t First initial and last name of the first author, add \'et al.\' if more authors are present",
                "{aA3etal}":" \t=\t First initial and last name of the first three authors (separated by comma), add \'et al.\' if more authors are present",
                "{T}":" \t\t=\t Title."}

valid_months = {'jan':'01','january':'01',
                'feb':'02','february':'02',
                'mar':'03','march':'03',
                'apr':'04','april':'04',
                'may':'05','may':'05',
                'jun':'06','june':'06',
                'jul':'07','july':'07',
                'aug':'08','august':'08',
                'sep':'09','september':'09',
                'oct':'10','october':'10',
                'nov':'11','november':'11',
                'dec':'12','december':'12'}

def month_to_number(month_string):
    return valid_months.get(month_string,"00")

def validate_journal(journal):
    return journal

def is_valid_integer(string,number_digits):
    return (string.isnumeric() and len(string)==number_digits)

def replace_bad_characters(string):
    replace ={'{\\\'{a}}'       :   'a',      
              '{\\~{n}}'        :   'n',     
              ':'               :   ' - ',        
              '{\\textendash}'  :   '-',       
              "?"               :   ".",  
              "{\\textemdash}"  :   "-",
              "\n"              :   " "}
    for i, j in replace.items():
        string = string.replace(i, j)

    invalid = "<>\"/\|*{}'"
    for char in invalid:
        string = string.replace(char, '')

    return string

def find_abbreviation_journal(journal_name):
    to_search = replace_bad_characters( (journal_name.strip() + " = ").lower() )
    if config.additional_abbreviations_file:
        with open(config.additional_abbreviations_file, 'r') as file:
            for line in file:
                if (line.lower()).startswith(to_search):
                    return line[len(to_search):].rstrip()

    data = pkgutil.get_data(__name__, "journalList.txt").decode('utf8')
    #with open('journalList.txt', 'r') as file:
    for line in data.splitlines():
        if (line.lower()).startswith(to_search):
            return line[len(to_search):].rstrip()
    return None

def find_tags_in_format(format):
    tags = re.findall(r'\{.*?\}', format)    #Create a list of all the tags used in this format, by looking for stuff wrapped between { and }
                                        # Example, res = ['{YYYY}', '{MM}', '{J}', '{A}', '{T}']
    return tags


def build_filename(infos,format,tags,max_length_authors=config.max_length_authors, max_length_filename=config.max_length_filename):

    #Based on the format specified by the user, we initialize a dictionary where
    #all the keys correspond to the tags used in the specified format (i.e. contained in the format string)
    #and the values are initially set to None

    rep_dict =  dict.fromkeys(tags)          #Initialize a dictionary with keys equal to the elements of the list tags, and all the values set to None


    #Now we look in the keys of the rep_dict dictionary and populate the values of the dictionary rep_dict by using the information contained in the infos dictionary
   
    if '{YYYY}' in rep_dict.keys():
        rep_dict['{YYYY}'] = infos['year'] if ('year' in infos and is_valid_integer(infos['year'],4)) else '0000'

    if '{MM}'in rep_dict.keys():
        rep_dict['{MM}'] = '00'
        if ('month' in infos):
            if is_valid_integer(infos['month'],2):
                rep_dict['{MM}'] = infos['month']
            elif is_valid_integer(infos['month'],1):
                rep_dict['{MM}'] = '0'+infos['month']
            elif (infos['month'].lower() in valid_months.keys()):
                rep_dict['{MM}'] = month_to_number(infos['month'].lower())

    if '{DD}'in rep_dict.keys():
        rep_dict['{DD}'] = '00'
        if ('day' in infos):
            if is_valid_integer(infos['day'],2):
                rep_dict['{DD}'] = infos['day']
            elif is_valid_integer(infos['day'],1):
                rep_dict['{DD}'] = '0'+infos['day']

    if ('{J}' in rep_dict.keys()) or ('{Jabbr}' in rep_dict.keys()):
        if ('journal' in infos) and infos['journal']:
            rep_dict['{J}'] = validate_journal(infos['journal'])
            Jabbr = find_abbreviation_journal(infos['journal'])
            if Jabbr:
                rep_dict['{Jabbr}'] = Jabbr
            else:
                rep_dict['{Jabbr}'] = rep_dict['{J}']
        elif ('ejournal' in infos) and infos['ejournal']:
            rep_dict['{J}'] = infos['ejournal']
            rep_dict['{Jabbr}'] = infos['ejournal']
        else:
            rep_dict['{J}'] = '[NoJournal]'
            rep_dict['{Jabbr}'] = '[NoJourn]'


    ListAuthorTags = ['{Aall}','{A3etal}','{Aetal}','{aAall}','{aA3etal}','{aAetal}']
    if any(item in rep_dict.keys() for item in ListAuthorTags):
        if ('author' in infos) and infos['author']:
            authors = [author.strip() for author in infos['author'].split(" and ")]
            lastnames = [name.split()[-1] for name in authors]
            firstnames = [name.split()[:-1] if len(name.split())>1 else [''] for name in authors]   #   The check on len(name.split())>1 is necessary to address the case in which
            if lastnames:                                                                                        #   the string name contains only one words (e.g. only the last name of the author is available)
                rep_dict['{Aall}'] = ", ".join(lastnames)
                rep_dict['{A3etal}'] = ", ".join(lastnames[0:3])
                if len(lastnames)>3:
                    rep_dict['{A3etal}'] = rep_dict['{A3etal}'] + " et al."
                rep_dict['{Aetal}'] = lastnames[0] + (" et al." if len(lastnames)>1 else "")
            else:
                for tag in ListAuthorTags:
                    rep_dict[tag] = '[NoAuthor]'

            if firstnames: 
                firstinitials = [firstname[0][0].upper()+"."  if len(firstname[0])>0 else "" for firstname in firstnames]
                firstinitial_lastnames = [firstinitials + " " + lastname for (firstinitials,lastname) in zip(firstinitials,lastnames) ]
  
                rep_dict['{aAall}'] = ", ".join(firstinitial_lastnames)
                rep_dict['{aA3etal}'] = ", ".join(firstinitial_lastnames[0:3])
                if len(firstinitial_lastnames)>3:
                    rep_dict['{aA3etal}'] = rep_dict['{aA3etal}'] + " et al."
                rep_dict['{aAetal}'] = firstinitial_lastnames[0] + (" et al." if len(firstinitial_lastnames)>1 else "")
            else:
                rep_dict['{aAall}'] = rep_dict['{Aall}']
                rep_dict['{aA3etal}'] = rep_dict['{A3etal}']
                rep_dict['{aAetal}'] = rep_dict['{Aetal}']
        else:
            for tag in ListAuthorTags:
                rep_dict[tag] = '[NoAuthor]'

        #Check that none of the author strings is longer than max_length_authors. If they are, we truncate it
        for tag in ListAuthorTags:
            if tag in rep_dict.keys():
                rep_dict[tag] = rep_dict[tag][0:max_length_authors]


    if '{T}' in rep_dict.keys():
        if ('title' in infos) and infos['title']:
            rep_dict['{T}'] = infos['title']
        else:
            rep_dict['{T}'] = '[NoTitle]'

    for key in rep_dict.keys():
        format = format.replace(key, rep_dict[key])
    filename = replace_bad_characters(format)
    #Check that the filename string is not longer than max_length_filename, and truncate it in case. 
    filename = filename[0:max_length_filename]
    return filename


import platform
system = platform.system()

if system.lower() == ('windows'): 
    separator = '\\'
else: 
    separator = '/'

# Setup logging
import logging
logger = logging.getLogger("pdf-renamer")
logger.setLevel(level=logging.INFO)
if not logger.handlers:
    formatter = logging.Formatter("[pdf-renamer] %(message)s")
    ch = logging.StreamHandler()
    ch.setFormatter(formatter)
    logger.addHandler(ch)
logger.propagate = False

#Parameters for this library
format = "{YYYY}-{MM} - {J} - {A3etal} - {T}" #Default value for the filename format, which will be used if no format is specified by user
additional_abbreviations_file = None #Global variable, it stores the path of a potential additional file with journal abbreviations
max_length_authors = 80 #Maximum length of any string related to authors
max_length_filename = 250 #Maximum filename length

#Parameters for the pdf2doi library
check_online_to_validate = True
websearch = True
numb_results_google_search = 6 #How many results should it look into when doing a google search